﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Drawing;
using System.Windows.Forms;
using System.IO;
using System.Collections;

namespace UrlCleaner
{
    /// <summary>
    /// Class to manage icons.
    /// 
    /// Author: Dennis Lang - 2011
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    class ExtractIcon
    {
        public ExtractIcon()
        {
            // Populate file type and icon table.
            iconsInfo = RegisteredFileType.GetFileTypeAndIcon();
        }

        // Registered file type table.
        Hashtable iconsInfo;   
 

        // Cache of icon extension's  image index.
        Dictionary<string, int> extIcons = new Dictionary<string, int>();

        /// <summary>
        /// Get small and large icon associated with file extension or folder
        /// </summary>
        /// <param name="name"></param>
        /// <param name="folderIcon">true=get Folder or link icon</param>
        /// <param name="smallIcons"></param>
        /// <param name="largeIcons"></param>
        /// <returns></returns>
        public int GetIcon(string name, bool folderIcon, ImageList smallIcons, ImageList largeIcons)
        {
            int iconIndex;
            string ext = Path.GetExtension(name).ToLower();

            if (ext.Length > 0)
            {
                if (ext == ".ico" || ext == ".png")
                {
                    try
                    {
                        Image image = Bitmap.FromFile(name);
                        return AddImage(image, smallIcons, largeIcons);
                    }
                    catch { }
                }

                // Check if icon already loaded. Don't cache exe or lnk 
                if (!folderIcon && ext != ".exe" && ext != ".lnk"
                    && this.extIcons.TryGetValue(ext, out iconIndex))
                    return iconIndex;

                // Get registered file icon
                iconIndex = GetRegIcon(ext, smallIcons, largeIcons);
                if (iconIndex != -1)
                    return iconIndex;
            }
        
            // Get shell file icon
            return GetShellIcon(name, folderIcon, smallIcons, largeIcons);
        }


        /// <summary>
        ///  Get registered icons.
        /// </summary>
        /// <param name="ext"></param>
        /// <param name="smallIcons"></param>
        /// <param name="largeIcons"></param>
        /// <returns></returns>
        public int GetRegIcon(string ext, ImageList smallIcons, ImageList largeIcons)
        {
            try
            {
                string fileAndParam = (this.iconsInfo[ext]).ToString();

                if (String.IsNullOrEmpty(fileAndParam))
                    return -1;

                int iconIdx = RegisteredFileType.ExtractIconFromFile(fileAndParam, smallIcons, largeIcons);
                if (iconIdx != -1 && ext != ".exe" && ext != ".lnk")
                    this.extIcons.Add(ext, iconIdx);

                return iconIdx;
            }
            catch (Exception ex)
            {
                // throw ex;
            }

            return -1;
        }


        /// <summary>
        /// Gets the Shell Icon for the given file...
        /// </summary>
        /// <param name="name">Path to file.</param>
        /// <param name="linkOverlay">Link Overlay</param>
        /// <returns>Icon</returns>
        public int GetShellIcon(string name, bool folderIcon, ImageList smallImages, ImageList largeImages)
        {
            int iconIndex;
            string ext = Path.GetExtension(name);

            // Check if icon already loaded. Don't cache exe or lnk 
            if (!folderIcon && ext != ".exe" && ext != ".lnk" 
                && this.extIcons.TryGetValue(ext, out iconIndex))
                return iconIndex;

            // Icon ico1 = Icon.ExtractAssociatedIcon(name);

            Shell32.SHFILEINFO shfi = new Shell32.SHFILEINFO();
            uint flags = Shell32.SHGFI_ICON | Shell32.SHGFI_USEFILEATTRIBUTES;
            uint dwAttr = Shell32.FILE_ATTRIBUTE_NORMAL;

            if (folderIcon)
            {
                flags = Shell32.SHGFI_ICON | Shell32.SHGFI_OPENICON;
                // flags += Shell32.SHGFI_LINKOVERLAY;
                // dwAttr = Shell32.FILE_ATTRIBUTE_DIRECTORY;
                dwAttr = 0;
            }
            if (ext == ".lnk") 
                flags += Shell32.SHGFI_LINKOVERLAY;

            Shell32.SHGetFileInfo(name,
                dwAttr,
                ref shfi,
                (uint)System.Runtime.InteropServices.Marshal.SizeOf(shfi),
                flags + Shell32.SHGFI_SMALLICON);

            // Check if we already have folder icon.
            int iconKey = 0;
            if (folderIcon)
            {
                iconKey = shfi.iIcon;
                if (folderImageIndex.TryGetValue(iconKey, out iconIndex))
                    return iconIndex;
            }

            // Clone  returned icon, Destroy original Icon
            Icon smallIcon = (System.Drawing.Icon)System.Drawing.Icon.FromHandle(shfi.hIcon).Clone();
            User32.DestroyIcon(shfi.hIcon);

            Shell32.SHGetFileInfo(name,
                dwAttr,
                ref shfi,
                (uint)System.Runtime.InteropServices.Marshal.SizeOf(shfi),
                flags + Shell32.SHGFI_LARGEICON);

            // Clone  returned icon, Destroy original Icon
            Icon largeIcon = (System.Drawing.Icon)System.Drawing.Icon.FromHandle(shfi.hIcon).Clone();
            User32.DestroyIcon(shfi.hIcon);

            // Add new image to lists
            iconIndex = smallImages.Images.Count;
            smallImages.Images.Add(smallIcon);
            largeImages.Images.Add(largeIcon);

            // Update image index database
            if (!folderIcon && ext != ".exe" && ext != ".lnk")
                this.extIcons.Add(ext, iconIndex);
            else if (folderIcon)
                folderImageIndex.Add(iconKey, iconIndex);

            return iconIndex;
        }

        /// <summary>
        /// Add image to small and large image list. Remove duplicates. Return image index position.
        /// </summary>
        /// <param name="new Image to add"></param>
        /// <param name="small Image list"></param>
        /// <param name="large Image list"></param>
        /// <returns></returns>
        public int AddImage(Image newImage, ImageList smallImages, ImageList largeImages)
        {
            int imageIdx = largeImages.Images.Count;
            largeImages.Images.Add(newImage);   // Add image so it gets resized before we compare it.

            // Check for duplicate image.
            for (int idx = 0; idx != imageIdx; idx++)
            {
                if (largeImages.Images[idx] == largeImages.Images[imageIdx])
                {
                    // Duplicate - remove added image.
                    largeImages.Images.RemoveAt(imageIdx);
                    return idx;
                }
            }

            // Need to add new image.
            smallImages.Images.Add(newImage);
            return imageIdx;
        }

        Dictionary<int, int> folderImageIndex = new Dictionary<int, int>();
    }
}
