﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Collections;
using System.Diagnostics;

namespace UrlCleaner 
{
    unsafe class ImageUtil
    {
        private struct ColorFreq
        {
            public Int32 count;
            public Int32 color;
        }

        private class SortColorFreq : IComparer
        {
            int IComparer.Compare(Object x, Object y)
            {
                // Sort reverse order so highest freq is first.
                ColorFreq cf1 = (ColorFreq)x;
                ColorFreq cf2 = (ColorFreq)y;
                return (cf2.count - cf1.count);
            }
        }

        private class SortColorBrightness : IComparer
        {
            int IComparer.Compare(Object x, Object y)
            {
                Color c1 = (Color)x;
                Color c2 = (Color)y;
                return c1.GetBrightness().CompareTo(c2.GetBrightness());
            }
        }

        private class SortColorRGB : IComparer
        {
            int IComparer.Compare(Object x, Object y)
            {
                Color c1 = (Color)x;
                Color c2 = (Color)y;
                uint v1 = RgbValue(c1);
                uint v2 = RgbValue(c2);
                return (v1 < v2) ? -1 : (v1 == v2) ? 0 : 1;
            }

            private uint RgbValue(Color c)
            {
                if (Math.Abs((int)c.R - (int)c.G) < 4 &&
                    Math.Abs((int)c.R - (int)c.B) < 4)
                {
                    // gray
                    return c.R;
                }
                else
                    if (c.R >= c.G && c.R >= c.B)
                        return c.R * 256u * 256u * 256u + c.G * 256u + c.B;
                    else if (c.G >= c.R && c.G >= c.B)
                        return c.G * 256u * 256u + c.B;
                    else
                        return c.B * 256u;
            }
        }

        static private Color[] SortColors(Color[] colors)
        {
            ArrayList cList = new ArrayList(colors);
            cList.Sort(new SortColorRGB());
            Color[] sorted = new Color[colors.Length];
            int n = 0;
            foreach (Color c in cList)
                sorted[n++] = c;
            return sorted;
        }

                       
        static byte BestColorIndex(Color[] colors, Color c32)
        {
            int bestIdx = 0;
            int minDist = int.MaxValue;

            for (int clrIdx = 0; clrIdx < colors.Length; clrIdx++)
            {
                Color cs = colors[clrIdx];
                int dist = Math.Abs(cs.R - c32.R) + Math.Abs(cs.G - c32.G) + Math.Abs(cs.B - c32.B);
                if (dist < minDist)
                {
                    minDist = dist;
                    bestIdx = clrIdx;
                }
            }

            System.Diagnostics.Debug.WriteLine(bestIdx);
            return (byte)bestIdx;
        }

        /// <summary>
        /// Extract a reasonable color Palette from a Truecolor image
        /// </summary>
        /// <param name="inImage">Input truecolor image</param>
        /// <returns>color array extracted from image</returns>
        static public Color[] Palette(Bitmap inImage, bool alwaysBuildColors)
        {
            int height = inImage.Height;
            int width = inImage.Width;
            Rectangle bounds = new Rectangle(0, 0, width, height);
            ColorPalette colorPalette = inImage.Palette;
            BitmapData sourceData = null;
            Color[] colors = colorPalette.Entries;

            if (colors.Length == 0 || alwaysBuildColors)
            {
                // Okay - we failed to get a quick color palette, time to make one
                try
                {
                    // Get the source image bits and lock into memory
                    sourceData = inImage.LockBits(bounds, ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                    byte* pSourceRow = (byte*)sourceData.Scan0.ToPointer();
                    Int32* pSourcePixel;
                    Hashtable colorMap = new Hashtable();
                    Int32 zero = 0;
                    Int32 maskARGB = 0x00ffffff; // remove alpha.

                    // Count colors present in image.
                    for (int row = 0; row < height; row++)
                    {
                        pSourcePixel = (Int32*)pSourceRow;

                        for (int col = 0; col < width; col++, pSourcePixel++)
                        {
                            Int32 c = *pSourcePixel & maskARGB;
                            if (colorMap.Contains(c))
                                colorMap[c] = (Int32)colorMap[c] + 1;
                            else
                            {
                                colorMap.Add(c, zero);
                                if (colorMap.Count == 4096)
                                {
                                    maskARGB = 0x00fcfcfc;
                                }
                                else if (colorMap.Count == 1024)
                                {
                                    maskARGB = 0x00fefefe;
                                }
                            }
                        }

                        pSourceRow += sourceData.Stride;
                    }

                    // Transfer Hash to Array so we can sort by count.
                    int nColors = colorMap.Keys.Count;
                    ArrayList sortFreq = new ArrayList();
                    foreach (Int32 n in colorMap.Keys)
                    {
                        ColorFreq colorFreq;
                        colorFreq.color = (Int32)n;
                        colorFreq.count = (Int32)colorMap[n];
                        if (colorFreq.count > 0)
                            sortFreq.Add(colorFreq);
                    }

                    sortFreq.Sort(new SortColorFreq());

                    // Add colors to list, keeping only those which are not 
                    //   too similar.  Adding from sorted list (most common to
                    //   less common), until 256 colors added.
                    colors = new Color[256];
                    Int32 colorsCnt = 1;    // Add first (most common color)
                    colors[0] = Color.FromArgb(((ColorFreq)sortFreq[0]).color);

                    Int32 minDist = 16;     // similar colors ignored (dist < 16)
                    Int32 srtCnt = sortFreq.Count;
                    for (int srtIdx = 1; srtIdx < srtCnt && colorsCnt < 256; srtIdx++)
                    {
                        Int32 cARGB = ((ColorFreq)sortFreq[srtIdx]).color;
                        int r = (byte)(cARGB >> 16);
                        int g = (byte)(cARGB >> 8);
                        int b = (byte)(cARGB);
                        Int32 d = 0;
                        for (int ic = 0; ic < colorsCnt; ic++)
                        {
                            Color cs = colors[ic];
                            d = Math.Abs(r - cs.R) + Math.Abs(g - cs.G) + Math.Abs(b - cs.B);
                            if (d < minDist)
                                break;
                        }

                        if (d > minDist)
                        {
                            // Store dissimiliar color and force alpha to 255
                            Color ca = Color.FromArgb(cARGB);
                            colors[colorsCnt++] = Color.FromArgb(255, ca.R, ca.G, ca.B);
                        }
                    }

#if false
                    // Sort for fun.
                    colors = SortColors(colors);
#endif
                    colorMap = null;
                }
                finally
                {
                    // Ensure that the bits are unlocked
                    inImage.UnlockBits(sourceData);
                }
            }

            return colors;
        }

        static public Bitmap Create8BitImage(Bitmap inImage, bool alwaysBuildColors)
        {
            Color[] colors = Palette(inImage, alwaysBuildColors);

            Rectangle bounds = new Rectangle(0, 0, inImage.Width, inImage.Height);
            Bitmap outImage8 = new Bitmap(inImage.Width, inImage.Height, PixelFormat.Format8bppIndexed);

            ColorPalette palette = outImage8.Palette;
            for (int clrIdx = 0; clrIdx < 256; clrIdx++)
            {
                palette.Entries[clrIdx] = colors[clrIdx];
            }
            outImage8.Palette = palette;

            BitmapData srcData = null;
            BitmapData dstData = null;
            try
            {
                srcData = inImage.LockBits(bounds, ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                byte* pSrcRow = (byte*)srcData.Scan0.ToPointer();
                Int32* pSrcPixel;
                Int32 maskARGB = 0x00ffffff; // remove alpha.

                dstData = outImage8.LockBits(bounds, ImageLockMode.WriteOnly, PixelFormat.Format8bppIndexed);
                byte* pDstRow = (byte*)dstData.Scan0.ToPointer();
                byte* pDstPixel;

                for (int row = 0; row < inImage.Height; row++)
                {
                    pSrcPixel = (Int32*)pSrcRow;
                    pDstPixel = (byte*)pDstRow;

                    for (int col = 0; col < inImage.Width; col++)
                    {
                        Int32 c32 = pSrcPixel[col] & maskARGB;
                        pDstPixel[col] = BestColorIndex(colors, Color.FromArgb(c32));
                    }

                    pSrcRow += srcData.Stride;
                    pDstRow += dstData.Stride;
                }
            }
            finally
            {
                // Ensure that the bits are unlocked
                inImage.UnlockBits(srcData);
                outImage8.UnlockBits(dstData);
            }

            return outImage8;
        }
    }
}
