﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Media;
using System.Text.RegularExpressions;
using System.Net;
using System.Net.NetworkInformation;
using System.Diagnostics;

/// <summary>
/// URL cleaner 
/// 
/// Features:
/// * scan directory for files with extension .url and populate sortable list view.
/// * verify all or selected files by loading resource
/// * set checkbox if load fails
/// * use can delete or move failed URL files.
/// 
/// Author: Dennis Lang - 2011
/// http://home.comcast.net/~lang.dennis/
/// </summary>

/*
Example of Url files:
 
    [{000214A0-0000-0000-C000-000000000046}]
    Prop3=19,0
    [InternetShortcut]
    URL=steam://rungameid/34330
    IDList=
    IconFile=C:\Program Files (x86)\Steam\steam\games\71a76cd2fbcd1457887fe57727aa10e5e1ce2ea4.ico
    IconIndex=0

 Another example:
    [InternetShortcut]
    URL=http://technet.microsoft.com/en-us/library/cc723564.aspx
 
*/
namespace UrlCleaner
{
    /// <summary>
    /// URL cleaner main class and user interface.
    /// 
    /// Author: Dennis Lang - 2011
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public partial class UrlCleaner : Form
    {
        public UrlCleaner()
        {
            InitializeComponent();

            // Constants to Listview columns.
            colnDir = colDir.Index;
            colnName = colName.Index;
            colnDate = colDate.Index;
            colnUrl = colURL.Index;             // tag is string, url to web image.
            colnStatus = colStatus.Index;
            colnOrgIcon = colOrgIcon.Index;     // tag is integer, to large image offset.
            colnWebIcon = colWebIcon.Index;     // tag is integer, to large image offset.
            colnNumImage = colImages.Index;
            colnNumLink = colLinks.Index;
            colnNumPage = colPages.Index;
            colnNumBlankPage = colBlankPages.Index;
            colnDuration = colDuration.Index;

            // Make both image lists the same size.
            m_largeImages.ImageSize = new System.Drawing.Size(48, 48);
            foreach (Image image in smallImages.Images)
            {
                m_largeImages.Images.Add(image);
            }

            findDialog.findEventHandler += new FindDialog.FindEventHandler(findDialog_findEventHandler);
	
            this.Text = Application.ProductName + " v" + Application.ProductVersion;
            urlListView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAuto);
            ResetBrowser();
            this.pathTx.Text = System.Environment.GetFolderPath(System.Environment.SpecialFolder.Desktop);
            this.timeoutTx.Text = m_timeout.Seconds.ToString();

            // Restore previously saved settings.
            RestoreSettings();

            // Default some controls off (not visible).
            verifyAbortGrp.Visible = false;
            delMovGrp.Visible = false;
        }

        #region ---- Data Members ----
        public class FindInfo
        {
            public int index;
        }
        FindInfo findInfo = new FindInfo() { index = 0 };
        FindDialog findDialog = new FindDialog();

        ListViewItem m_hoverItem = null;
        ListViewItem m_activeItem = null;

        bool m_docDone = false;
        bool m_ignorePageDone = false;
        uint m_goodPageCnt = 0;
        uint m_blankPageCnt = 0;
        Stopwatch m_stopWatch = new Stopwatch();
        List<string> m_logList = new List<string>();

        uint m_goodCnt = 0;
        uint m_badCnt = 0;
        const string GoodBrowserStatusStr = "Done";
        const string AbortBrowserStatusStr = "Aborted";

        // Constants to column positions.
        int colnDir = 0;
        int colnName = 1;
        int colnDate = 2;
        int colnUrl = 3;
        int colnStatus = 4;
        int colnOrgIcon = 5;
        int colnWebIcon = 6;
        int colnNumImage = 7;
        int colnNumLink = 8;
        int colnNumPage = 9;
        int colnNumBlankPage = 10;
        int colnDuration = 11;

        int emptyImageIdx = 4;

        ImageList m_largeImages = new ImageList();
        int m_webIconImageIdx = -1;

        OptionsDlg m_optionsDlg = null;
        SoundPlayer m_sndPlayer = new SoundPlayer();
        bool m_mute = false;

        TimeSpan m_timeout = TimeSpan.FromSeconds(10);

        FilterUI.CheckedValue[] m_includeFilters = new FilterUI.CheckedValue[] 
        { 
            new FilterUI.CheckedValue() { enable=true, value="http://" },
            new FilterUI.CheckedValue() { enable=true, value="https://" },
        };
        FilterUI.CheckedValue[] m_excludeFilters = new FilterUI.CheckedValue[]
        {
            new FilterUI.CheckedValue() { enable=true, value="stream://" },
            new FilterUI.CheckedValue() { enable=true, value=".pdf" },
        };

        #endregion 
       
        #region ---- Basic UI controls ----
        /// <summary>
        /// Save MediaMap and icon path when application closes.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
        {
            SaveSettings();
            base.OnClosing(e);
        }

        /// <summary>
        /// User closing program via 'close' button.
        /// </summary>
        private void closeBtn_Click(object sender, EventArgs e)
        {
            Close();
        }

        AboutDlg m_aboutDlg;
        private void aboutBtn_Click(object sender, EventArgs e)
        {
            m_aboutDlg = new AboutDlg();
            m_aboutDlg.Show();
        }

        #endregion

        #region ---- Information Bar button controls
        /// <summary>
        /// Get new Timeout value (in seconds).
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void timeoutTx_TextChanged(object sender, EventArgs e)
        {
            uint timeoutSeconds;

            if (uint.TryParse(timeoutTx.Text, out timeoutSeconds))
            {
                m_timeout = TimeSpan.FromSeconds(timeoutSeconds);
            }
        }

        /// <summary>
        ///  Save web icon locally as eitehr Png or 32bit Icon
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void saveIconBtn_Click(object sender, EventArgs e)
        {
            if (m_webIconImageIdx != -1)
            {
                if (saveIconDlg.ShowDialog() == DialogResult.OK)
                {
                    try
                    {
                        if (saveIconDlg.FilterIndex == 1)
                        {
                            Bitmap image32 = new Bitmap(m_largeImages.Images[m_webIconImageIdx], new Size(48, 48));
#if true
                            System.IO.FileStream saveToStream =
                                    new System.IO.FileStream(saveIconDlg.FileName, System.IO.FileMode.OpenOrCreate);
                            // Save 32bit icon.
                            IconUtil.SaveBitmapAsIcon(image32, saveToStream);
#else
                            Bitmap image8 = ImageUtil.Create8BitImage(image32, true);
                            image8.Save(saveIconDlg.FileName, System.Drawing.Imaging.ImageFormat.Icon);

                            Icon icon = Icon.FromHandle(image8.GetHicon());
                            image8.Dispose();
                            System.IO.FileStream saveToStream = 
                                    new System.IO.FileStream(saveIconDlg.FileName, System.IO.FileMode.OpenOrCreate);
                            // This only saves a 4bit icon (16 colors).
                            icon.Save(saveToStream);
                            icon.Dispose();
#endif
                        }
                        else
                        {
                            Bitmap image = new Bitmap(m_largeImages.Images[m_webIconImageIdx]);
                            image.Save(saveIconDlg.FileName);
                        }
                    }
                    catch (Exception ex)
                    {
                        this.statusTx.Text = ex.Message;
                    }
                }
            }
            else
            {
                this.statusTx.Text = "No icon available";
            }
        }

        /// <summary>
        ///  Save Icon locally and Set URL to use icon for presentation.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void setIconBtn_Click(object sender, EventArgs e)
        {
            if (m_activeItem != null)
            {
                try
                {
                    string path = Path.Combine(m_activeItem.SubItems[colnDir].Text, m_activeItem.SubItems[colnName].Text);
                    IniFile iniFile = new IniFile(path);

                    /*
                    Example of Url files:
 
                        [{000214A0-0000-0000-C000-000000000046}]
                        Prop3=19,0
                        [InternetShortcut]
                        URL=steam://rungameid/34330
                        IDList=
                        IconFile=C:\Program Files (x86)\Steam\steam\games\71a76cd2fbcd1457887fe57727aa10e5e1ce2ea4.ico
                        IconIndex=0

                     Another example:
                        [InternetShortcut]
                        URL=http://technet.microsoft.com/en-us/library/cc723564.aspx
 
                    */

                    string urlStr = iniFile.Read("InternetShortcut", "URL");
                    if (urlStr.Length != 0)
                    {
                        Uri uri = new Uri(urlStr);

                        string iconPath = System.Environment.GetFolderPath(Environment.SpecialFolder.MyPictures);
                        iconPath = Path.Combine(iconPath, uri.Host.Replace('.', '_'));
                        iconPath += ".ico";

                        int webIconImageIdx = (int)m_activeItem.SubItems[colnWebIcon].Tag;
                        Bitmap image32 = new Bitmap(m_largeImages.Images[webIconImageIdx], new Size(48, 48));
                        System.IO.FileStream saveToStream =
                                    new System.IO.FileStream(iconPath, System.IO.FileMode.OpenOrCreate);
                        IconUtil.SaveBitmapAsIcon(image32, saveToStream);
                        saveToStream.Close();

                        iniFile.Write("InternetShortcut", "IconFile", iconPath);
                        iniFile.Write("InternetShortcut", "IconIndex", "0");
                        iniFile.Write("InternetShortcut", "IDList", "");
                        iniFile.Write("InternetShortcut", "HotKey", "0");
                    }
                }
                catch (Exception ex)
                {
                    this.statusTx.Text = ex.Message;
                }

            }
        }

        /// <summary>
        /// Set URL to use web icon image directly from network.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void setWebBtn_Click(object sender, EventArgs e)
        {
            if (m_activeItem != null)
            {
                try
                {
                    string urlWebImage = (string)m_activeItem.SubItems[colnUrl].Tag;
                    string path = Path.Combine(m_activeItem.SubItems[colnDir].Text, m_activeItem.SubItems[colnName].Text);
                    IniFile iniFile = new IniFile(path);

                    string urlStr = iniFile.Read("InternetShortcut", "URL");
                    if (urlStr.Length != 0)
                    {
                        /*
                        Example URL with icon set directly to web page.
             
                        [InternetShortcut]
                        URL=http://stackoverflow.com/questions/871234/circular-lock-free-buffer
                        IconFile=http://sstatic.net/stackoverflow/img/favicon.ico
                        IconIndex=1
                        */
                        iniFile.Write("InternetShortcut", "IconFile", urlWebImage);
                        iniFile.Write("InternetShortcut", "IconIndex", "1");
                    }
                }
                catch (Exception ex)
                {
                    this.statusTx.Text = ex.Message;
                }
            }
        }

        #endregion

        #region ---- Log file ----
        private string LogFile()
        {
            return System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + @"\UrlCleaner.log";
        }
        private void LogIt(string msg)
        {
            try
            {
                StreamWriter stream = File.AppendText(LogFile());
                stream.WriteLine(DateTime.Now.ToString("G") + ", " + msg);
            }
            catch { }
        }
        #endregion

        #region ---- Save and Restore settings ----
        /// <summary>
        /// Return path to persistent settings file.
        /// </summary>
        /// <returns></returns>
        private string SettingsFile()
        {
            return System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + @"\UrlCleaner.dat";
        }
   
        /// <summary>
        /// Restore saved MediaMap path and icon path.
        /// </summary>
        private void RestoreSettings()
        {
            try
            {
                StreamReader stream = File.OpenText(SettingsFile());
                string lineString;
                while ((lineString = stream.ReadLine()) != null)
                {
                    lineString = lineString.Trim();
                    string[] fields = lineString.Split('=');
                    if (fields.Length == 2 && lineString[0] != ';')
                    {
                        switch (fields[0].ToLower())
                        {
                            case "path":
                                pathTx.Text = fields[1].Trim();
                                break;
                            case "include":
                                m_includeFilters = FilterUI.Parse(fields[1].Trim());
                                break;
                            case "exclude":
                                m_excludeFilters = FilterUI.Parse(fields[1].Trim());
                                break;
                        }
                    }
                }
                stream.Close();
            }
            catch
            { }
        }

        /// <summary>
        /// Save some user values so they can be restored next time application starts.
        /// </summary>
        private void SaveSettings()
        {
            try
            {
                StreamWriter stream = File.CreateText(SettingsFile());

                stream.WriteLine("path=" + pathTx.Text);
                stream.WriteLine("include=" + FilterUI.ToString(m_includeFilters));
                stream.WriteLine("exclude=" + FilterUI.ToString(m_excludeFilters));
                stream.Close();
            }
            catch
            { }
        }
        #endregion

        #region ---- URL Path controls ----
        private void pathTx_TextChanged(object sender, EventArgs e)
        {
            scanBtn.Enabled = (pathTx.Text.Length != 0) /* && Directory.Exists(pathTx.Text) */;
        }
      
        private void pathTx_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (e.KeyCode == Keys.Tab)
                e.IsInputKey = true;
        }

        private void pathBtn_Click(object sender, EventArgs e)
        {
            try
            {
                string[] dirs = pathTx.Text.Split(';');

                if (Directory.Exists(dirs[0]))
                {
                    folderDlg.SelectedPath = dirs[0];
                    folderDlg.RootFolder = System.Environment.SpecialFolder.Desktop;
                }

            }
            catch (Exception ex)
            {
                this.statusTx.Text = ex.Message;
            }

            if (folderDlg.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                pathTx.Text = folderDlg.SelectedPath;
            }
        }
        #endregion

        #region ---- URL Scan and Filter ----
        List<Regex> m_includeRegex = new List<Regex>();
        List<Regex> m_excludeRegex = new List<Regex>();

        private bool HasInclude(string urlStr)
        {
            foreach (Regex regex in m_includeRegex)
            {
                if (regex.IsMatch(urlStr))
                    return true;
            }
            return m_includeRegex.Count == 0;
        }
        private bool NotExclude(string urlStr)
        {
            foreach (Regex regex in m_excludeRegex)
            {
                if (regex.IsMatch(urlStr))
                    return false;
            }
            return true;
        }

        private void scanBtn_Click(object sender, EventArgs e)
        {
            m_includeRegex.Clear();
            foreach (FilterUI.CheckedValue ckValue in m_includeFilters)
            {
                if (ckValue.enable)
                    m_includeRegex.Add(new Regex(ckValue.value));
            }
            m_excludeRegex.Clear();
            foreach (FilterUI.CheckedValue ckValue in m_excludeFilters)
            {
                if (ckValue.enable)
                    m_excludeRegex.Add(new Regex(ckValue.value));
            }

            this.urlListView.Items.Clear();
            this.urlListView.ListViewItemSorter = null;
            this.urlListView.BeginUpdate();
            this.Cursor = Cursors.WaitCursor;
            this.progressBar.Visible = true;
            this.progressBar.Value = 0;
            this.progressBar.Style = ProgressBarStyle.Marquee;
            this.progressBar.MarqueeAnimationSpeed = 100;
            this.verifyAbortGrp.Visible = true;      // enable abort button.
            this.scanBtn.Enabled = false;

            this.m_abort = false;
            this.m_goodCnt = this.m_badCnt = 0;

            string[] dirs = pathTx.Text.Split(';');

            foreach (string dir in dirs)
            {
                if (Directory.Exists(dir))
                    Scan(new DirectoryInfo(dir));
            }

            this.Text = Application.ProductName;
            this.statusTx.Text = string.Format("{0} URL's found", urlListView.Items.Count);
            this.verifyAbortGrp.Visible = (urlListView.Items.Count != 0);
            this.scanBtn.Enabled = true;


            this.progressBar.Visible = false;
            this.Cursor = Cursors.Default;
            this.urlListView.EndUpdate();
            this.urlListView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAuto);
        }

        private void Scan(DirectoryInfo dirInfo)
        {
            if (this.m_abort)
            {
                statusTx.Text = "Scan Aborted";
                return;
            }

            // Allow ui events to execute (this is dangerous and should be avoided).
            // Better solution is to scan directory in a background thread.
            Application.DoEvents();

            try
            {
                foreach (DirectoryInfo subDirInfo in dirInfo.GetDirectories())
                {
                    Scan(subDirInfo);
                }
            }
            catch (Exception ex)
            {
                this.statusTx.Text = ex.Message;
                return;
            }

            foreach (FileInfo urlInfo in dirInfo.GetFiles("*.url"))
            {
                try
                {
                    ListViewItem lvItem = new ListViewItem(urlInfo.Directory.FullName);
                    lvItem.SubItems.Add(urlInfo.Name);
                    lvItem.SubItems.Add(urlInfo.LastWriteTime.ToString("G"));

                    IniFile iniFile = new IniFile(urlInfo.FullName);
                    string urlStr = iniFile.Read("InternetShortcut", "URL");
                    if (urlStr.Length != 0 && HasInclude(urlStr) && NotExclude(urlStr))
                    {
                        lvItem.SubItems.Add(urlStr);
                        urlListView.Items.Add(lvItem);
                        this.Text = string.Format("{0} {1}", urlListView.Items.Count, Application.ProductName);
                        this.viewCount.Text = urlListView.Items.Count.ToString("#,##0");

                        int iconIdx = -1;
                        string iconPath = iniFile.Read("InternetShortcut", "IconFile");
                        Uri uri = new Uri(urlStr);
                        if (iconPath.Length != 0 && (!uri.IsFile || !File.Exists(iconPath)) && Uri.IsWellFormedUriString(iconPath, UriKind.RelativeOrAbsolute))
                        {
                            // ToDo - queue action to background thread
                            // ToDo - check and resolve relative Url
                            Image iconImage;
                            if (RetrieveUrlImage(iconPath, out iconImage))
                                iconIdx = extractIcon.AddImage(iconImage, smallImages, m_largeImages);
                            else
                                iconIdx = extractIcon.GetIcon("foo.html", false, smallImages, m_largeImages);
                        }
                        else
                        {
                            if (iconPath.Length == 0 || !File.Exists(iconPath))
                                iconPath = "foo.html";

                            // int iconIdx = extractIcon.GetIcon(urlInfo.FullName, false, smallImages, largeImages);
                            iconIdx = extractIcon.GetIcon(iconPath, false, smallImages, m_largeImages);
                        }

                        while (lvItem.SubItems.Count < colnWebIcon + 1)
                            lvItem.SubItems.Add(string.Empty);
                        SetImage(lvItem, colnOrgIcon, iconIdx);
                    }
                }
                catch (Exception ex)
                {
                    this.statusTx.Text = ex.Message;
                }
            }
        }
        #endregion

        #region ---- URL Listview methods ----
        private void urlListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            this.m_abort = false;
            Point pt = urlListView.PointToClient(System.Windows.Forms.Control.MousePosition);
            ListViewHitTestInfo hitInfo = urlListView.HitTest(pt);
            if (hitInfo.Item != null)
            {
                hitInfo.Item.Checked = !hitInfo.Item.Checked;   // undo checked box change on double click.
                Verify(hitInfo.Item);
            }
        }

        /// <summary>
        /// Column header click fires sort.
        /// Supports two tier sorting.
        /// </summary>
        private void ColumnSort_Click(object sender, ColumnClickEventArgs e)
        {
            bool isControl = System.Windows.Forms.Control.ModifierKeys == Keys.Control;

            ListView listView = sender as ListView;
            ListViewColumnSorter sorter = listView.ListViewItemSorter as ListViewColumnSorter;
            if (sorter == null)
                return;

#if false
            if (listView.Groups.Count != 0)
            {
                // MessageBox.Show("Sorting does not work when in 'Group' mode");
                foreach (ListViewGroup lvGroup in listView.Groups)
                    SortGroup(lvGroup);
            }
#endif

            // Determine if clicked column is already the column that is being sorted.
            if (e.Column == sorter.SortColumn1)
            {
                // Reverse the current sort direction for this column.
                if (sorter.Order1 == SortOrder.Ascending)
                    sorter.Order1 = SortOrder.Descending;
                else
                    sorter.Order1 = SortOrder.Ascending;
            }
            else
            {
                // Set the column number that is to be sorted; default to ascending.
                sorter.SortColumn2 = sorter.SortColumn1;
                sorter.Order2 = sorter.Order1;
                sorter.SortColumn1 = e.Column;
                sorter.Order1 = SortOrder.Ascending;
            }

            // Clear old arrows and set new arrow
            foreach (ColumnHeader colHdr in listView.Columns)
                colHdr.ImageIndex = 4;

            if (sorter.SortColumn1 != sorter.SortColumn2)
                listView.Columns[sorter.SortColumn2].ImageIndex = (sorter.Order2 == SortOrder.Ascending) ? 2 : 3;

            listView.Columns[sorter.SortColumn1].ImageIndex = (sorter.Order1 == SortOrder.Ascending) ? 0 : 1;

            sorter.SortCheckedState = (isControl && sorter.SortColumn1 == 0);

            sorter.SortType =
                (listView.Columns[sorter.SortColumn1].TextAlign == HorizontalAlignment.Right) ?
                ListViewColumnSorter.SortDataType.eNumeric :
                ListViewColumnSorter.SortDataType.eAlpha;

            // Perform the sort with these new sort options.
            if (listView != null)
                listView.Sort();
        }

        private void urlListView_MouseMove(object sender, MouseEventArgs e)
        {
            int selCnt = urlListView.SelectedItems.Count;
            if (e.Button == MouseButtons.Right && selCnt != 0)
            {
                string[] fileList = new string[selCnt];
                for (int idx = 0; idx != selCnt; idx++)
                {
                    ListViewItem lvItem = urlListView.SelectedItems[idx];
                    fileList[idx] = Path.Combine(lvItem.SubItems[colnDir].Text, lvItem.SubItems[colnName].Text);
                }

                this.DoDragDrop(new DataObject(DataFormats.FileDrop, fileList), DragDropEffects.Copy);
            }
        }

        private void urlListView_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Delete:
                    break;
              
                case Keys.V:    // verify (load) url link.
                    verifyLoadToolStripMenuItem_Click(null, EventArgs.Empty);
                    break;
            }
        }

        private void UrlCleaner_KeyUp(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.F:    // Launch Find dialog in response to control-f.
                    if (e.Control)
                        findMenu_Click(sender, EventArgs.Empty);
                    break;
                case Keys.F3:
                    Find(urlListView, findInfo, new FindDialog.FindEventArgs() { action = FindDialog.FindEventArgs.Action.Next });
                    break;
            }
        }

        private void urlListView_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            delMovGrp.Visible = (urlListView.CheckedItems.Count != 0);
        }

        private void urlListView_ItemMouseHover(object sender, ListViewItemMouseHoverEventArgs e)
        {
            m_hoverItem = e.Item;
            ShowUrlDetail(m_hoverItem);
        }

        /// <summary>
        /// Set status message and line color
        /// </summary>
        /// <param name="lvItem"></param>
        /// <param name="msg"></param>
        /// <returns>true if status changed</returns>
        private bool SetUriStatus(ListViewItem lvItem, string msg)
        {
            while (lvItem.SubItems.Count < colnStatus + 1)
                lvItem.SubItems.Add("");
            if (lvItem.SubItems[colnStatus].Text != msg)
            {
                lvItem.SubItems[colnStatus].Text = msg;
                if (msg == GoodBrowserStatusStr)
                {
                    overlay.Visible = false;
                    UpdateGood(1);
                    PlaySound(Sounds.eGood);
                    lvItem.BackColor = Color.PaleGreen;
                }
                else if (msg == AbortBrowserStatusStr)
                {
                    overlay.Visible = true;
                    lvItem.BackColor = Color.LightYellow;
                    lvItem.Checked = false;
                }
                else
                {
                    overlay.Visible = true;
                    UpdateBad(1);
                    PlaySound(Sounds.eFail);
                    lvItem.BackColor = Color.LightPink;

                    lvItem.Checked = true;
                }

                return true;
            }

            return false;
        }

        private void SetUriDetails(ListViewItem lvItem, WebBrowser browser, 
            uint goodPageCnt, uint blankPageCnt, TimeSpan duration)
        {
            while (lvItem.SubItems.Count < colnDuration + 1)
                lvItem.SubItems.Add("");

            int imageCnt = 0;
            if (goodPageCnt > 1)
                try { imageCnt = (int)lvItem.SubItems[colnNumImage].Tag; }
                catch { }
            imageCnt += browser.Document.Images.Count;
            lvItem.SubItems[colnNumImage].Text = imageCnt.ToString("#,##0");
            lvItem.SubItems[colnNumImage].Tag = imageCnt;

            int linkCnt = 0;
            if (goodPageCnt > 1) 
                try { linkCnt = (int)lvItem.SubItems[colnNumLink].Tag; } 
                catch {}
            linkCnt += browser.Document.Links.Count;
            lvItem.SubItems[colnNumLink].Text = linkCnt.ToString("#,##0");
            lvItem.SubItems[colnNumLink].Tag = linkCnt;

            lvItem.SubItems[colnNumPage].Text = goodPageCnt.ToString("#,##0");
            lvItem.SubItems[colnNumBlankPage].Text = blankPageCnt.ToString("#,##0");
            lvItem.SubItems[colnDuration].Text = duration.Seconds.ToString("#,##0");

            int imageIdx = GetImage(lvItem, colnOrgIcon);
            orgIcon.BackgroundImage = (imageIdx != -1) ? m_largeImages.Images[imageIdx] : null;

            Image urlImage;
            string urlPath;
            m_webIconImageIdx = -1;

            // ToDo - check and resolve relative Url
            if (!browser.IsBusy)
            {
                if (RetrieveUrlImage(out urlImage, out urlPath))
                {
                    imageIdx = extractIcon.AddImage(urlImage, smallImages, m_largeImages);

                    while (lvItem.SubItems.Count < colnWebIcon + 1)
                        lvItem.SubItems.Add(string.Empty);
                    SetImage(lvItem, colnWebIcon, imageIdx);

                    webIcon.BackgroundImage = m_largeImages.Images[imageIdx];
                    m_webIconImageIdx = imageIdx;
                    lvItem.SubItems[colnUrl].Tag = urlPath;
                }
                else if (browser.Document.Images.Count != 0)
                {
                    string imageUrl = browser.Document.Images[0].GetAttribute("src");
                    System.Diagnostics.Debug.WriteLine("Image: " + imageUrl);
                    if (RetrieveUrlImage(imageUrl, out urlImage))
                    {
                        imageIdx = extractIcon.AddImage(urlImage, smallImages, m_largeImages);

                        while (lvItem.SubItems.Count < colnWebIcon + 1)
                            lvItem.SubItems.Add(string.Empty);
                        SetImage(lvItem, colnWebIcon, imageIdx);

                        webIcon.BackgroundImage = m_largeImages.Images[imageIdx];
                        m_webIconImageIdx = imageIdx;
                    }
                }
            }

            if (m_webIconImageIdx == -1)
            {
                webIcon.BackgroundImage = null;
                SetImage(lvItem, colnWebIcon, emptyImageIdx);
            }
        }
        #endregion

        #region ---- Image/Icon methods ----
        ExtractIcon extractIcon = new ExtractIcon();
        OAKListView.LV_ITEM oakLvItem;
 
        private void SetImage(ListViewItem lvItem, int colIdx, int imageIdx)
        {
            oakLvItem = new OAKListView.LV_ITEM();
            // Row
            oakLvItem.iItem = lvItem.Index;
            // Column
            oakLvItem.iSubItem = colIdx;
            if (lvItem.SubItems[colIdx].Text.Length != 0)
            {
                oakLvItem.pszText = lvItem.SubItems[colIdx].Text;
                oakLvItem.mask = OAKListView.LVIF_IMAGE | OAKListView.LVIF_TEXT;
            }
            else
            {
                oakLvItem.mask = OAKListView.LVIF_IMAGE;
            }
            // Image index on imagelist
            oakLvItem.iImage = imageIdx;
            lvItem.SubItems[colIdx].Tag = imageIdx;
            OAKListView.SendMessage(lvItem.ListView.Handle, OAKListView.LVM_SETITEM, 0, ref oakLvItem);
            Application.DoEvents();
        }

        private int GetImage(ListViewItem lvItem, int colIdx)
        {
            try
            {
                if (lvItem.SubItems.Count > colIdx)
                    return (int)lvItem.SubItems[colIdx].Tag;
            }
            catch { }

            return -1;
        }

        private bool RetrieveUrlImage(string imageUrl, out Image image)
        {
            bool gotImage = false;
            image = null;
            imageUrl = RemoveHtmlEscapeCodes(imageUrl);

            System.Diagnostics.Stopwatch stopwatch = new System.Diagnostics.Stopwatch();
            string webStatus = string.Empty;
            stopwatch.Start();

            HttpWebResponse response = null;

            try
            {
                HttpWebRequest request = (HttpWebRequest)HttpWebRequest.Create(imageUrl);

                request.Timeout = m_timeout.Milliseconds;
                request.ReadWriteTimeout = m_timeout.Milliseconds;
                request.AllowAutoRedirect = false;
                response = (HttpWebResponse)request.GetResponse();

                if (response.StatusCode != HttpStatusCode.OK)
                {
                    webStatus = response.StatusDescription;
                }

                string baseURL = GetBaseURL(response);
                string contentType = response.ContentType;
                string encoding = response.ContentEncoding;

                if (contentType.Length == 0)
                {
                    response.Close();
                    return false;
                }

                System.Diagnostics.Debug.WriteLine("Page content type: " + contentType);

                // if (contentType.Contains("image/"))
                {
                    Stream resStream = response.GetResponseStream();
                    System.IO.Stream webDataStream = resStream;

                    try
                    {
                        image = System.Drawing.Image.FromStream(webDataStream);
                        gotImage = true;
                    }
                    catch (Exception ex)
                    {
                        statusTx.Text = ex.Message + " " + imageUrl;
                    }

                    response.Close();
                }
            }
            catch (Exception ex)
            {
                statusTx.Text = ex.Message + " " + imageUrl;
            }

            return gotImage;
        }

        /// <summary>
        /// Extract "Shortcut Icon" from web document.
        /// </summary>
        /// <param name="image">Return Icon image</param>
        /// <param name="uriPath">Return Icon URI</param>
        /// <returns></returns>
        private bool RetrieveUrlImage(out Image image, out string uriPath)
        {
            uriPath = string.Empty;

            try
            {
                HtmlDocument doc = webBrowser.Document;
                HtmlElementCollection collect = doc.GetElementsByTagName("link");

                foreach (HtmlElement element in collect)
                {
                    string linkRelStr = element.GetAttribute("rel");

                    if (string.Compare(linkRelStr, "SHORTCUT ICON", true) == 0 ||
                        string.Compare(linkRelStr, "ICON", true) == 0)
                    {
                        string iconPath = element.GetAttribute("href");
                        uriPath = iconPath;

                        if (Uri.IsWellFormedUriString(uriPath, UriKind.Relative))
                            uriPath = "http://" + UriCombine(webBrowser.Url.Host, iconPath);

                        WebRequest requestImg = WebRequest.Create(uriPath);
                        image = Image.FromStream(requestImg.GetResponse().GetResponseStream());
                        return true;
                    }
                }
            }
            catch (Exception ex)
            {
                statusTx.Text = ex.Message + " " + uriPath;
            }

            image = null;
            return false;
        }
        #endregion

        #region ---- URL Verification methods ----
        private void verifyAllBtn_Click(object sender, EventArgs e)
        {
            this.m_abort = false;
            this.progressBar.Visible = true;
            this.progressBar.Value = 0;
            this.progressBar.Style = ProgressBarStyle.Continuous;

            this.m_goodCnt = this.m_badCnt = 0;
            // this.urlListView.BeginUpdate();
            this.Cursor = Cursors.WaitCursor;

            foreach (ListViewItem lvItem in this.urlListView.Items)
            {
                if (lvItem.SubItems.Count > colnStatus)
                    lvItem.SubItems[colnStatus].Text = string.Empty;

                Verify(lvItem);
                progressBar.Value = (int)(lvItem.Index * 100 / urlListView.Items.Count);
            }

            this.progressBar.Visible = false;
            this.Cursor = Cursors.Default;
            // this.urlListView.EndUpdate();
        }

        private void verifySelBtn_Click(object sender, EventArgs e)
        {
            this.m_abort = false;
            this.progressBar.Visible = true;
            this.progressBar.Value = 0;
            // this.urlListView.BeginUpdate();
            this.Cursor = Cursors.WaitCursor;

            int cnt = 0;
            foreach (ListViewItem lvItem in this.urlListView.SelectedItems)
            {
                Verify(lvItem);
                cnt++;
                progressBar.Value = (int)(cnt * 100 / urlListView.SelectedItems.Count);
            }

            this.progressBar.Visible = false;
            this.Cursor = Cursors.Default;
            // this.urlListView.EndUpdate();
        }

        enum Sounds { eNone, eGood, eFail, eAbort };
        DateTime m_lastPlay = DateTime.Now;
        /// <summary>
        /// Play a sound.  Ignore play if recently played.
        /// </summary>
        /// <param name="sounds"></param>
        private void PlaySound(Sounds sounds)
        {
            if (!this.m_mute)
            {
                this.m_sndPlayer.Stream = Properties.Resources.bad;
                switch (sounds)
                {
                    case Sounds.eAbort:
                    case Sounds.eFail:
                        this.m_sndPlayer.Stream = Properties.Resources.bad;
                        break;
                    case Sounds.eGood:
                        this.m_sndPlayer.Stream = Properties.Resources.good;
                        break;
                }

                if ((DateTime.Now - m_lastPlay).Duration().TotalSeconds > 2)
                {
                    this.m_sndPlayer.Play();
                    m_lastPlay = DateTime.Now;
                }
            }
        }

        private void muteBtn_Click(object sender, EventArgs e)
        {
            this.m_mute = !this.m_mute;
            muteBtn.BackgroundImage = this.m_mute ? 
                global::UrlCleaner.Properties.Resources.muteOn :
                global::UrlCleaner.Properties.Resources.muteOff; 
        }

        /// <summary>
        /// Ping host - verify connection.
        /// </summary>
        /// <param name="host">Host name or ip string.</param>
        /// <param name="addressList"></param>
        /// <returns>True if successful</returns>
        private bool Pingable(string host, IPAddress[] addressList)
        {
            // Set the ping options, TTL 128
            PingOptions pingOptions = new PingOptions(128, true);

            // 32 byte buffer (create empty)
            byte[] buffer = new byte[32];
            int timeout = 600;

            Ping ping = new Ping();
            PingReply reply = ping.Send(host, timeout, buffer, pingOptions);
            return (reply.Status == IPStatus.Success);
        }

        /// <summary>
        /// Verify url entry, load url resource.
        ///     * Well formed url 
        ///     * Well formed hostname
        ///     * DNS hostname lookup
        ///     * browser loads one or more pages, status is "Done"
        ///       Error if "blank" page loaded.
        /// </summary>
        /// <param name="lvItem"></param>
        private void Verify(ListViewItem lvItem)
        {
            m_activeItem = lvItem;
            string urlStr = lvItem.SubItems[colnUrl].Text;

            if (this.m_abort)
            {
                SetUriStatus(lvItem, AbortBrowserStatusStr);
                return;
            }

            while (lvItem.SubItems.Count < colnDuration + 1)
                lvItem.SubItems.Add("");
            for (int idx = colnStatus; idx <= colnDuration; idx++)
            {
                lvItem.SubItems[idx].Text = string.Empty;
                lvItem.SubItems[idx].Tag = null;
            }

            try
            {
                Uri uri = new Uri(urlStr);

                if (!Uri.CheckSchemeName(uri.Scheme))
                {
                    SetUriStatus(lvItem, "Invalid Scheme " + uri.Scheme);
                }
                else if (Uri.CheckHostName(uri.Host) == UriHostNameType.Unknown)
                {
                    SetUriStatus(lvItem, "Unknown Host " + uri.Host);
                }
                else if (!uri.IsWellFormedOriginalString())
                {
                    SetUriStatus(lvItem, "Invalid uri " + urlStr);
                }
                else
                {
                    string host = uri.DnsSafeHost;
                    IPAddress[] addressList = Dns.GetHostAddresses(host);
                    if (addressList == null)
                    {
                        SetUriStatus(lvItem, "Unknown Host " + host);
                        return;
                    }
#if false

                    if (!Pingable(host, addressList))
                    {
                        SetUriStatus(lvItem, "ping failed to " + host);
                        return;
                    }
#endif

                    System.Diagnostics.Debug.WriteLine("Load " + uri.OriginalString);
                    m_goodPageCnt = 0;
                    m_blankPageCnt = 0;
                    m_stopWatch.Reset();
                    m_stopWatch.Start();

                    if (!LoadWebPage(uri))
                    {
                        // ResetBrowser(); // Stop browser
                        this.webBrowser.Stop();
                        m_docDone = false;
                        m_ignorePageDone = false;
                        SetUriStatus(lvItem, "Timed Out");
                        m_badCnt++;
                    }
                    
                }
            }
            catch (Exception ex)
            {
                SetUriStatus(lvItem, ex.Message);
            }
        }

        private void UpdateGood(int dir)
        {
            if (dir > 0)
                m_goodCnt++;
            else if (m_goodCnt > 0)
                m_goodCnt--;
            goodTx.Text = m_goodCnt.ToString();
        }

        private void UpdateBad(int dir)
        {
            if (dir > 0)
                m_badCnt++;
            else if (m_badCnt > 0)
                m_badCnt--;
            badTx.Text = m_badCnt.ToString();
        }
        #endregion 

        #region ---- WEB/URL resource methods ----
        /// <summary>
        /// 
        /// </summary>
        void webBrowser_DocumentCompleted(object sender, WebBrowserDocumentCompletedEventArgs e)
        {
            // System.Diagnostics.Debug.WriteLine("Complete status: " + webBrowser.StatusText + " " + e.Url.OriginalString);

            if (m_docDone)
                return;     // Already handled page load, ignore duplicate events.

            string loadStatus = webBrowser.StatusText;
            if (e.Url.OriginalString == null || e.Url.OriginalString == "about:blank")
            {
                m_blankPageCnt++;
                if (m_blankPageCnt >= m_goodPageCnt)
                    loadStatus = "Blank Page";
#if false
                m_docDone = true;
                m_ignorePageDone = true;
                webBrowser.Stop();
#endif
            }
            else
                m_goodPageCnt++;    // count non-blank pages loaded.

            this.statusTx.Text = loadStatus;

            if (m_activeItem != null /* && m_activeItem.SubItems[colnStatus].Text == webBrowser.Url.ToString() */)
            {
#if false
                string urlStr = m_activeItem.SubItems[colnUrl].Text;
                if (urlStr != webBrowser.Url.OriginalString && m_goodPageCnt == 1)
                {
                    loadStatus = "Redirected to " + webBrowser.Url.Host;
                }
#endif

                string percent = this.progressBar.Visible ? string.Format("{0}% ", progressBar.Value) : string.Empty;
                this.statusTx.Text = string.Format("{0} Item #{1} Verification Status:{2}, Url:{3}",
                    percent,
                    m_activeItem.Index, loadStatus, m_activeItem.SubItems[colnUrl].Text);

                this.Text = percent + Application.ProductName;
#if false
                if (SetUriStatus(m_activeItem, webBrowser.StatusText))
                {
                    // Status changed - update details.
                    SetUriDetails(m_activeItem, webBrowser);
                }
                else
                {
                    // Status has not change - display old details.
                    ShowUrlDetail(m_activeItem);
                }
#else
                SetUriStatus(m_activeItem, loadStatus);
                SetUriDetails(m_activeItem, webBrowser, m_goodPageCnt, m_blankPageCnt, m_stopWatch.Elapsed);
#endif
                if (this.autoScrollCk.Checked)
                    m_activeItem.EnsureVisible();
            }

            m_docDone = !webBrowser.IsBusy;
            m_logList.Add(string.Format("{0} {1}/{2} {3}", m_docDone.ToString(), m_goodPageCnt, m_blankPageCnt, webBrowser.Url.ToString()));
        }

        static string RemoveHtmlEscapeCodes(string rawString)
        {
            // ToDo - add additional escape sequences.
            return rawString.Replace("&amp;", "&");
        }

        static string GetBaseURL(HttpWebResponse response)
        {
            string fullURI = response.ResponseUri.AbsoluteUri;
            string baseURL = fullURI.Substring(0, fullURI.IndexOf(response.ResponseUri.AbsolutePath));
            return baseURL;
        }

        private string UriCombine(string host, string path)
        {
            return host + ((path[0] == '.') ? path.Substring(1) : path);
        }

        bool browserInit = false;
        private void ResetBrowser()
        {
            // webBrowser.Url = null;
            webBrowser.Stop();

            this.mainSplit.Panel2.Controls.Remove(webBrowser);
            webBrowser = new WebBrowser();
            this.mainSplit.Panel2.Controls.Add(webBrowser);
            webBrowser.Dock = DockStyle.Fill;
          
        //    if (!browserInit)
            {
                browserInit = true;
                webBrowser.DocumentCompleted += new WebBrowserDocumentCompletedEventHandler(webBrowser_DocumentCompleted);
               
                webBrowser.ScriptErrorsSuppressed = true;

                // Subscribe to Event(s) with the WindowsInterop Class
                WindowsInterop.SecurityAlertDialogWillBeShown +=
                    new GenericDelegate<Boolean, Boolean>(this.WindowsInterop_SecurityAlertDialogWillBeShown);

                WindowsInterop.ConnectToDialogWillBeShown +=
                    new GenericDelegate<String, String, Boolean>(this.WindowsInterop_ConnectToDialogWillBeShown);
            }
        }

        private Boolean WindowsInterop_SecurityAlertDialogWillBeShown(Boolean blnIsSSLDialog)
        {
            // Return true to ignore and not show the 
            // "Security Alert" dialog to the user
            return true;
        }

        private Boolean WindowsInterop_ConnectToDialogWillBeShown(ref String sUsername, ref String sPassword)
        {
            // (Fill in the blanks in order to be able 
            // to return the appropriate Username and Password)
            sUsername = "";
            sPassword = "";

            // Return true to auto populate credentials and not 
            // show the "Connect To ..." dialog to the user
            return true;
        }

        private bool m_abort = false;

        private bool LoadWebPage(Uri uri)
        {
            m_logList.Add("Load " + uri.ToString());

            m_docDone = false;
            m_ignorePageDone = false;
            ResetBrowser();
            overlay.Visible = false;
            webBrowser.Url = uri;
            DateTime startAt = DateTime.Now;
            TimeSpan maxWait = TimeSpan.FromSeconds(30);

            while (this.m_abort == false && m_docDone == false && (DateTime.Now - startAt) < maxWait)
            {
                // TODO - run in background using urlget and not browser widget.
                Application.DoEvents();
                // System.Threading.Thread.Sleep(100);
            }
            // Application.DoEvents();
            return m_docDone;
        }
        #endregion

        #region ---- Main Button click methods ----
        private void abortBtn_Click(object sender, EventArgs e)
        {
            this.m_abort = true;
            PlaySound(Sounds.eAbort);
        }

        /// <summary>
        /// Permentantly delete url file ! 
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void deleteBtn_Click(object sender, EventArgs e)
        {
            Flash(urlListView.CheckedItems, 2);
            if (MessageBox.Show(string.Format("Are you sure you want to delete {0} url shortcuts ?", urlListView.CheckedItems.Count),
                "Confirm Delete", MessageBoxButtons.YesNo) == System.Windows.Forms.DialogResult.Yes)
            {
                foreach (ListViewItem lvItem in urlListView.CheckedItems)
                {
                    DeleteFile(lvItem);
                }
            }
        }

        private void clearCntBtn_Click(object sender, EventArgs e)
        {
            m_goodCnt = m_badCnt = 0;
            goodTx.Text = badTx.Text = string.Empty;
        }

        private void clearColors_Click(object sender, EventArgs e)
        {
            urlListView.BeginUpdate();
            foreach (ListViewItem lvItem in urlListView.Items)
            {
                lvItem.BackColor = urlListView.BackColor;
            }
            urlListView.EndUpdate();
        }

        private void configBtn_Click(object sender, EventArgs e)
        {
            m_optionsDlg = new OptionsDlg();
            m_optionsDlg.Include = m_includeFilters;
            m_optionsDlg.Exclude = m_excludeFilters;

            if (m_optionsDlg.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                m_includeFilters = m_optionsDlg.Include;
                m_excludeFilters = m_optionsDlg.Exclude;
            }
        }

        private void copyBtn_Click(object sender, EventArgs e)
        {
            int selCnt = urlListView.CheckedItems.Count;
            if (selCnt != 0)
            {
                System.Collections.Specialized.StringCollection fileList = new System.Collections.Specialized.StringCollection();

                for (int idx = 0; idx != selCnt; idx++)
                {
                    ListViewItem lvItem = urlListView.CheckedItems[idx];
                    fileList.Add(Path.Combine(lvItem.SubItems[colnDir].Text, lvItem.SubItems[colnName].Text));
                }

                // this.DoDragDrop(new DataObject(DataFormats.FileDrop, fileList), DragDropEffects.Copy);
                Clipboard.SetText(Join("\r\n", fileList));
                // Clipboard.SetFileDropList(fileList);
            }
        }
        #endregion

        #region ---- Popup menu list methods -----
        private void openInBrowserToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_hoverItem != null)
            {
                System.Diagnostics.Process.Start(m_hoverItem.SubItems[colnUrl].Text);
            }
        }

        private void openContainerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_hoverItem != null)
            {
                string fullPath = Path.Combine(m_hoverItem.SubItems[colnDir].Text, m_hoverItem.SubItems[colnName].Text);
                if (File.Exists(fullPath))
                    System.Diagnostics.Process.Start("explorer.exe", "/select," + fullPath);
                else
                    statusTx.Text = "Invalid file " + fullPath;
            }
        }

        private void openInNotepadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_hoverItem != null)
            {
                string fullPath = Path.Combine(m_hoverItem.SubItems[colnDir].Text, m_hoverItem.SubItems[colnName].Text);
                if (File.Exists(fullPath))
                    System.Diagnostics.Process.Start("notepad.exe", fullPath);
                else
                    statusTx.Text = "Invalid file " + fullPath;
            }
        }

        private void verifyLoadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_hoverItem != null && urlListView.SelectedItems.Count != 0)
            {
                // Verify(m_hoverItem);
                foreach (ListViewItem lvItem in urlListView.SelectedItems)
                {
                    Verify(lvItem);
                }
                m_hoverItem = null;
            }
        }

        private void deleteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_hoverItem != null)
            {
                //  DeleteFile(m_hoverItem);
                Flash(urlListView.CheckedItems, 2);
                if (urlListView.SelectedItems.Count == 1 ||
                    MessageBox.Show(string.Format("Are you sure you want to delete {0} url shortcuts ?",
                    urlListView.SelectedItems.Count),
                    "Confirm Delete", MessageBoxButtons.YesNo) == System.Windows.Forms.DialogResult.Yes)
                {
                    foreach (ListViewItem lvItem in urlListView.SelectedItems)
                    {
                        DeleteFile(lvItem);
                    }
                    m_hoverItem = null;
                }
            }
        }

        private void removeFromListToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_hoverItem != null)
            {
                // this.urlListView.Items.Remove(m_hoverItem);
                foreach (ListViewItem lvItem in urlListView.SelectedItems)
                {
                    this.urlListView.Items.Remove(lvItem);
                }

                m_hoverItem = null;
            }
        }

        private void checkDuplicatesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            uint dupCnt = 0;

            urlListView.BeginUpdate();

            // Sort by Url column.
            ColumnSort_Click(urlListView, new ColumnClickEventArgs(colnUrl));

            int cnt = urlListView.Items.Count;
            if (cnt > 1)
            {
                string urlStr = urlListView.Items[0].SubItems[colnUrl].Text;
                for (int idx = 1; idx < cnt; idx++)
                {
                    urlListView.Items[idx].Checked = urlListView.Items[idx].SubItems[colnUrl].Text == urlStr;
                    if (urlListView.Items[idx].Checked)
                    {
                        dupCnt++;
                    }
                    else
                    {
                        urlStr = urlListView.Items[idx].SubItems[colnUrl].Text;
                    }
                }
            }

            urlListView.EndUpdate();
            statusTx.Text = string.Format("{0} Duplicates found in {1} items", dupCnt, cnt);
        }
        #endregion

        #region ---- Misc support methods ----
        private void Flash(ListView.CheckedListViewItemCollection items, int flashes)
        {
            while (flashes-- > 0)
            {
                foreach (ListViewItem lvItem in items)
                {
                    Color bgcolor = lvItem.BackColor;
                    lvItem.BackColor = lvItem.ForeColor;
                    lvItem.ForeColor = bgcolor;
                    lvItem.Selected = false;
                }

                Application.DoEvents();
                System.Threading.Thread.Sleep(1000);

                foreach (ListViewItem lvItem in items)
                {
                    Color bgcolor = lvItem.BackColor;
                    lvItem.BackColor = lvItem.ForeColor;
                    lvItem.ForeColor = bgcolor;
                }
                Application.DoEvents();
                System.Threading.Thread.Sleep(1000);
            }
        }

        private void DeleteFile(ListViewItem lvItem)
        {
            string fullPath = Path.Combine(lvItem.SubItems[colnDir].Text, lvItem.SubItems[colnName].Text);
            if (File.Exists(fullPath))
            {
                try
                {
                    RecycleFile.DeleteFile(fullPath);
                    // File.Delete(fullPath);
                    LogIt("Deleted " + fullPath);
                    if (lvItem.SubItems[colnStatus].Text == GoodBrowserStatusStr)
                        UpdateGood(-1);
                    else
                        UpdateBad(-1);

                    urlListView.Items.Remove(lvItem);
                }
                catch (Exception ex)
                {
                    statusTx.Text = ex.Message + " " + fullPath;
                }
            }
        }

        private void ShowUrlDetail(ListViewItem lvItem)
        {
            try
            {
                if (lvItem.SubItems.Count > colnStatus)
                {
                    string detail = string.Format("Path:{0}\\{1}\nLast Modified:{2}\nUrl:{3}",
                        lvItem.SubItems[colnDir].Text, lvItem.SubItems[colnName].Text,
                        lvItem.SubItems[colnDate].Text,
                        lvItem.SubItems[colnStatus].Text);
                }

                statusTx.Text = lvItem.SubItems[colnUrl].Text + " Status:" + lvItem.SubItems[colnStatus].Text;
                int imageIdx = GetImage(lvItem, colnOrgIcon);
                orgIcon.BackgroundImage = (imageIdx != -1) ? m_largeImages.Images[imageIdx] : null;
                imageIdx = GetImage(lvItem, colnWebIcon);
                webIcon.BackgroundImage = (imageIdx != -1) ? m_largeImages.Images[imageIdx] : null;
                m_webIconImageIdx = imageIdx;
            }
            catch (Exception ex)
            {
                this.statusTx.Text = ex.Message;
            }
        }
     
        private string Join(string separator, System.Collections.Specialized.StringCollection strList)
        {
            string result = string.Empty;
            foreach (string str in strList)
            {
                if (result.Length != 0)
                    result += separator;
                result += str;
            }
            return result;
        }
        #endregion

        #region ---- Find ----
        private void findMenu_Click(object sender, EventArgs e)
        {
            findDialog.GetArgs.action = FindDialog.FindEventArgs.Action.First;
            findDialog.Show();
        }

        private void findAgainMenu_Click(object sender, EventArgs e)
        {
            findDialog.GetArgs.action = FindDialog.FindEventArgs.Action.First;
            if (findDialog.GetArgs.toFind.Length == 0)
                findDialog.Show();
            else
                findDialog_findEventHandler(null, findDialog.GetArgs);
        }

        void findDialog_findEventHandler(object sender, FindDialog.FindEventArgs e)
        {
            Find(this.urlListView, findInfo, e);
        }

        void Find(ListView listView, FindInfo findInfo, FindDialog.FindEventArgs e)
        {
            if (e.action == FindDialog.FindEventArgs.Action.First)
            {
                listView.SelectedIndices.Clear();
                findInfo.index = -1;
            }
            else if (e.action == FindDialog.FindEventArgs.Action.All)
            {
                listView.SelectedIndices.Clear();
                findInfo.index = -1;
                while ((findInfo.index = FindMatch(this.urlListView, e, findInfo.index)) != -1)
                {
                    listView.EnsureVisible(findInfo.index);
                    listView.SelectedIndices.Add(findInfo.index);
                }

                if (listView.SelectedIndices.Count != 0)
                    PlaySound(Sounds.eGood);
                else
                    PlaySound(Sounds.eFail);

                return;
            }

            // Search ListView.
            findInfo.index = FindMatch(this.urlListView, e, findInfo.index);
            if (findInfo.index == -1)
            {
                MessageBox.Show("No Match for:" + e.toFind);
                PlaySound(Sounds.eFail);
            }
            else
            {
                PlaySound(Sounds.eGood);
                listView.EnsureVisible(findInfo.index);
                listView.SelectedIndices.Add(findInfo.index);
            }
        }

        public static bool Contains(string source, string toCheck, bool isMatchCase, bool isRegularExpression)
        {
            if (isRegularExpression)
                return Regex.IsMatch(source, toCheck, isMatchCase ? RegexOptions.None : RegexOptions.IgnoreCase);
            else
                return source.IndexOf(toCheck, isMatchCase ? StringComparison.Ordinal : StringComparison.OrdinalIgnoreCase) >= 0;
        }

        /// <summary>
        /// Find a match, starting search just after previous match.
        /// </summary>
        /// <param name="listView"></param>
        /// <param name="e"></param>
        /// <param name="prevIndex"></param>
        /// <returns>Matching row index or -1 on failure</returns>
        int FindMatch(ListView listView, FindDialog.FindEventArgs e, int prevIndex)
        {
            int index = prevIndex;
            while (++index < listView.Items.Count)
            {
                ListViewItem lvItem = listView.Items[index];
                for (int colIdx = 0; colIdx != lvItem.SubItems.Count; colIdx++)
                {
                    if (lvItem.SubItems[colIdx] != null && 
                        Contains(lvItem.SubItems[colIdx].Text, e.toFind, e.isMatchCase, e.isRegularExpression))
                        return index;
                }
            }

            return -1;
        }
        #endregion

        HelpDlg m_helpdlg;
        private void helpBtn_Click(object sender, EventArgs e)
        {
            this.m_helpdlg = new HelpDlg();
            this.m_helpdlg.Show();
        }

        UrlDlg m_urlDlg;
        private void urlBtn_Click(object sender, EventArgs e)
        {
            this.m_urlDlg = new UrlDlg();
            this.m_urlDlg.UrlList = this.m_logList;
            this.m_urlDlg.Show();
        }
    }
}
